import re
from typing import Type

from assimilator.core.database import BaseRepository, SpecificationList
from assimilator.core.database.exceptions import NotFoundError
from assimilator.internal.database.specifications import InternalSpecification, InternalSpecificationList


class InternalRepository(BaseRepository):
    def __init__(self, session: dict, specifications: Type[SpecificationList] = InternalSpecificationList):
        super(InternalRepository, self).__init__(session=session, initial_query='', specifications=specifications)

    def get(self, *specifications: InternalSpecification, lazy: bool = False):
        try:
            return self.session[self._apply_specifications(specifications)]
        except (KeyError, TypeError) as exc:
            raise NotFoundError(exc)

    def filter(self, *specifications: InternalSpecification, lazy: bool = False):
        if not specifications:
            return list(self.session.values())

        key_mask = self._apply_specifications(specifications)
        if lazy:
            return key_mask

        models = []
        for key, value in self.session.items():
            if not re.match(key, key_mask):
                pass

            models.append(value)

        return models

    def save(self, obj):
        self.session[str(obj.id)] = obj

    def delete(self, obj):
        del self.session[str(obj.id)]

    def update(self, obj):
        self.session[str(obj.id)] = obj

    def is_modified(self, obj):
        return self.get(self.specifications.filter(obj.id)) == obj

    def refresh(self, obj):
        obj.value = self.get(self.specifications.filter(obj.id))


__all__ = [
    'InternalRepository',
]



import json
from typing import Type

from pydantic import ValidationError, BaseModel

from assimilator.core.exceptions import ParsingError


class JSONParsedMixin:
    @classmethod
    def from_json(cls: Type['BaseModel'], data: str) -> 'BaseModel':
        try:
            return cls(**json.loads(data))
        except ValidationError as exc:
            raise ParsingError(exc)


__all__ = [
    'JSONParsedMixin',
]



from typing import Iterable

import redis

from assimilator.core.events import Event, ExternalEvent
from assimilator.core.events.events_bus import EventConsumer, EventProducer


class RedisEventConsumer(EventConsumer):
    def __init__(self, channels: Iterable[str], session: redis.Redis):
        self.session = session
        self.channels = channels

    def close(self):
        self.session.close()

    def start(self):
        """ Connected by default """

    def consume(self) -> Iterable[ExternalEvent]:
        publish_subscribe = self.session.pubsub()
        publish_subscribe.subscribe(*self.channels)

        while True:
            message = publish_subscribe.get_message()

            if (message is None) or message['type'] != 'message':
                self.delay_function()
                continue

            yield ExternalEvent.from_json(message['data'])
            self.delay_function()

    def delay_function(self):
        pass


class RedisEventProducer(EventProducer):
    def __init__(self, channel: str, session: redis.Redis):
        self.session = session
        self.channel = channel

    def produce(self, event: Event):
        self.session.publish(self.channel, event.json())

    def start(self):
        pass

    def close(self):
        self.close()


__all__ = [
    'RedisEventConsumer',
    'RedisEventProducer',
]



from typing import Optional

from pydantic import BaseModel

from assimilator.core.patterns.mixins import JSONParsedMixin


class RedisModel(JSONParsedMixin, BaseModel):
    id: int
    expire_in: Optional[int] = None


__all__ = [
    'RedisModel',
]



from typing import Type

from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Query

from assimilator.alchemy.database.specifications import AlchemySpecificationList
from assimilator.core.database import BaseRepository, Specification, SpecificationList
from assimilator.core.database.exceptions import NotFoundError


class AlchemyRepository(BaseRepository):
    def __init__(
        self, session, initial_query: Query = None, specifications: Type[SpecificationList] = AlchemySpecificationList,
    ):
        super(AlchemyRepository, self).__init__(
            session=session,
            initial_query=initial_query,
            specifications=specifications,
        )

    def _execute_query(self, query):
        return self.session.execute(query)

    def get(self, *specifications: Specification, lazy=False):
        try:
            data = self._execute_query(self._apply_specifications(specifications))
            if lazy:
                return data

            return data.one()[0]

        except NoResultFound as exc:
            raise NotFoundError(exc)

    def filter(self, *specifications: Specification, lazy=False):
        data = self._execute_query(self._apply_specifications(specifications))
        if lazy:
            return data

        return [result[0] for result in data]

    def update(self, obj):
        """ We don't do anything, as the object is going to be updated with the obj.key = value """

    def save(self, obj):
        self.session.add(obj)

    def refresh(self, obj):
        self.session.refresh(obj)

    def delete(self, obj):
        self.session.delete(obj)

    def is_modified(self, obj):
        return self.session.is_modified(obj)


__all__ = [
    'AlchemyRepository',
]



from assimilator.core.database import Specification, specification, SpecificationList


class InternalSpecification(Specification):
    def apply(self, query: str) -> str:     # returns the str key
        return super(InternalSpecification, self).apply(query)


@specification
def internal_filter(key: str, query: str) -> str:
    return f"{query}{key}"


@specification
def internal_order(*args, query: str, **kwargs) -> str:
    return query


@specification
def internal_paginate(*args, query: str, **kwargs) -> str:
    return query


@specification
def internal_join(*args, query: str, **kwargs) -> str:
    return query


class InternalSpecificationList(SpecificationList):
    filter = internal_filter
    order = internal_order
    paginate = internal_paginate
    join = internal_join


__all__ = [
    'InternalSpecification',
    'internal_filter',
    'internal_order',
    'internal_paginate',
    'internal_join',
    'InternalSpecificationList',
]



from typing import Type

import redis

from assimilator.core.database import SpecificationList
from assimilator.redis.database import RedisModel
from assimilator.core.database.repository import BaseRepository, LazyCommand
from assimilator.internal.database.specifications import InternalSpecification, InternalSpecificationList


class RedisRepository(BaseRepository):
    def __init__(
        self,
        session: redis.Redis,
        model: Type[RedisModel],
        specifications: Type[SpecificationList] = InternalSpecificationList,
    ):
        super(RedisRepository, self).__init__(session, initial_query='', specifications=specifications)
        self.model = model

    def get(self, *specifications: InternalSpecification, lazy: bool = False):
        key_name = self._apply_specifications(specifications)
        if lazy:
            return LazyCommand(self.session.get, key_name)
        return self.session.get(key_name)

    def filter(self, *specifications: InternalSpecification, lazy: bool = False):
        key_name = self._apply_specifications(specifications)
        if lazy:
            return LazyCommand(self.session.keys, key_name)

        return [self.model.from_json(value) for value in self.session.mget(self.session.keys(key_name))]

    def save(self, obj: RedisModel):
        self.session.set(str(obj.id), obj.json(), ex=obj.expire_in)

    def delete(self, obj: RedisModel):
        self.session.delete(str(obj.id))

    def update(self, obj: RedisModel):
        self.save(obj)

    def is_modified(self, obj: RedisModel):
        return self.get(self.specifications.filter(obj.id), lazy=False) == obj

    def refresh(self, obj: RedisModel):
        fresh_obj = self.get(self.specifications.filter(obj.id), lazy=False)

        for key, value in fresh_obj.dict().items():
            setattr(obj, key, value)


__all__ = [
    'RedisRepository',
]



from assimilator.core.exceptions import ParsingError


class EventError(Exception):
    pass


class EventParsingError(ParsingError, EventError):
    pass


class EventProducingError(EventError):
    pass


__all__ = [
    'EventError',
    'EventParsingError',
    'EventProducingError',
]



from typing import Iterable, Collection

from sqlalchemy.orm import Query

from assimilator.core.database.specifications import Specification, specification, SpecificationList


class AlchemySpecification(Specification):
    def apply(self, query: Query) -> Query:
        return super(AlchemySpecification, self).apply(query)


@specification
def alchemy_filter(*filters, query: Query, **filters_by) -> Query:
    return query.filter(*filters).filter_by(**filters_by)


@specification
def alchemy_order(*clauses: str, query: Query) -> Query:
    return query.order_by(*clauses)


@specification
def alchemy_paginate(limit: int, offset: int, query: Query) -> Query:
    return query.limit(limit).offset(offset)


@specification
def alchemy_join(targets: Collection, join_args: Collection[dict], query: Query) -> Query:
    if len(join_args) < len(targets):
        join_args += [dict() for _ in range(len(targets) - len(join_args))]

    for target, join_data in zip(targets, join_args):
        query = query.join(target, **join_data)

    return query


class AlchemySpecificationList(SpecificationList):
    filter = alchemy_filter
    order = alchemy_order
    paginate = alchemy_paginate
    join = alchemy_join


__all__ = [
    'AlchemySpecificationList',
    'AlchemySpecification',
    'alchemy_filter',
    'alchemy_order',
    'alchemy_paginate',
    'alchemy_join',
]



from abc import ABC, abstractmethod
from typing import Union, Any, Optional, Callable, Iterable, Type

from assimilator.core.database.specifications import SpecificationList, SpecificationType


class LazyCommand:
    def __init__(self, command: Callable, *args, **kwargs):
        self.command = command
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.command(*self.args, **self.kwargs)


class BaseRepository(ABC):
    def __init__(self, session: Any, specifications: Type[SpecificationList], initial_query: Optional[Any] = None):
        self.session = session
        self.initial_query = initial_query
        self.specifications = specifications

    def _get_initial_query(self):
        if self.initial_query is not None:
            return self.initial_query
        else:
            raise NotImplementedError("You must either pass the initial query or define get_initial_query()")

    def _apply_specifications(self, specifications: Iterable[SpecificationType]) -> Any:
        query = self._get_initial_query()

        for specification in specifications:
            query = specification(query)

        return query

    @abstractmethod
    def get(self, *specifications: SpecificationType, lazy: bool = False) -> Union[LazyCommand, Any]:
        raise NotImplementedError("get() is not implemented()")

    @abstractmethod
    def filter(self, *specifications: SpecificationType, lazy: bool = False) -> Union[LazyCommand, Iterable]:
        raise NotImplementedError("filter() is not implemented()")

    @abstractmethod
    def save(self, obj) -> None:
        raise NotImplementedError("save() is not implemented in the repository")

    @abstractmethod
    def delete(self, obj) -> None:
        raise NotImplementedError("delete() is not implemented in the repository")

    @abstractmethod
    def update(self, obj) -> None:
        raise NotImplementedError("update() is not implemented in the repository")

    @abstractmethod
    def is_modified(self, obj) -> bool:
        raise NotImplementedError("is_modified() is not implemented in the repository")

    @abstractmethod
    def refresh(self, obj) -> None:
        raise NotImplementedError("refresh() is not implemented in the repository")


__all__ = [
    'LazyCommand',
    'BaseRepository',
]



from typing import Type

from assimilator.core.database import UnitOfWork
from assimilator.core.services.base import Service


class CRUDService(Service):
    def __init__(self, uow: UnitOfWork, model: Type):
        self.uow = uow
        self.specifications = self.uow.repository.specifications
        self.model = model

    def create(self, obj_data: dict):
        with self.uow:
            obj = self.model(**obj_data)
            self.uow.repository.save(obj)
            self.uow.commit()

        self.uow.repository.refresh(obj)
        return obj

    def update(self, update_data: dict, *filters, **kwargs_filters):
        with self.uow:
            obj = self.get(*filters, **kwargs_filters)

            for key, value in update_data.items():
                setattr(obj, key, value)

            self.uow.repository.update(obj)
            self.uow.commit()

        self.uow.repository.refresh(obj)
        return obj

    def list(self, *filters, lazy: bool = False, **kwargs_filters):
        return self.uow.repository.get(self.specifications.filter(*filters, **kwargs_filters), lazy=lazy)

    def get(self, *filters, lazy: bool = False, **kwargs_filters):
        return self.uow.repository.filter(self.specifications.filter(*filters, **kwargs_filters), lazy=lazy)

    def delete(self, *filters, **kwargs_filters):
        with self.uow:
            obj = self.get(*filters, **kwargs_filters)
            self.uow.repository.delete(obj)
            self.uow.commit()


__all__ = [
    'CRUDService',
]



from abc import ABC, abstractmethod
from typing import Callable, Union
from functools import wraps


class Specification(ABC):
    @abstractmethod
    def apply(self, query):
        raise NotImplementedError("Specification must specify apply()")

    def __call__(self, query):
        return self.apply(query)


def specification(func: Callable):
    def create_specification(*args, **kwargs):

        @wraps(func)
        def created_specification(query):
            return func(*args, **kwargs, query=query)

        return created_specification
    return create_specification


SpecificationType = Union[Specification, Callable]


class SpecificationList:
    filter: SpecificationType
    order: SpecificationType
    paginate: SpecificationType
    join: SpecificationType


__all__ = [
    'SpecificationList',
    'Specification',
    'specification',
    'SpecificationType',
]
